// equalize.c
// makes a new RGB image by  
// heavily, and in a dirty(?) way, reuses code from resample.c 
// and on               image filter plug-in for The Gimp image
// manipulation program,  
// sorry, and thanks,
// Tue Jul 16 11:10:36 EDT 2002  <pac@next.gli.cas.cz>
// License: GPL
// Version 1.00




#include <u.h>
#include <libc.h>
#include <draw.h>
#include <memdraw.h>

#define ALPHA 3
static int     pos_lut[256];		/* Positive coefficient LUT */
static int     neg_lut[256];		/* Negative coefficient LUT */


void
usage(void)
{
	fprint(2, "usage:\n\tequalize [imagefile]\n");
	fprint(2, "\n\tjpg -3 testimg.jpg | equalize  | page  \n");
	fprint(2, "\n\tequalize  rose.img  | togif | gif\n");
	exits("usage");
}

/*********************************
 *
 *  blur_prepare_row()
 *
 *  Get a row of pixels.  If the requested row
 *  is off the edge, clone the edge row.
 *
 ********************************/

static  void
blur_prepare_row (GimpPixelRgn *pixel_rgn, 
		  guchar    *data, 
		  int        x, 
		  int        y, 
		  int        w)
{
  int b;

  if (y == 0)
    {
      gimp_pixel_rgn_get_row(pixel_rgn, data, x, (y + 1), w);
    }
  else if (y == pixel_rgn->h)
    {
      gimp_pixel_rgn_get_row(pixel_rgn, data, x, (y - 1), w);
    }
  else
    {
      gimp_pixel_rgn_get_row(pixel_rgn, data, x, y, w);
    }
  /*
   *  Fill in edge pixels
   */
  for (b = 0; b < pixel_rgn->bpp; b++)
    {
      data[-(gint)pixel_rgn->bpp + b] = data[b];
      data[w * pixel_rgn->bpp + b] = data[(w - 1) * pixel_rgn->bpp + b];
    }
}

/*********************************
 *
 *  blur()
 *
 *  Actually mess with the image.
 *
 ********************************/

static void
blur (GimpDrawable *drawable)
{
  GimpPixelRgn srcPR, destPR, destPR2, *sp, *dp, *tp;
  gint width, height;
  gint bytes;
  guchar *dest, *d;
  guchar *prev_row, *pr;
  guchar *cur_row, *cr;
  guchar *next_row, *nr;
  guchar *tmp;
  gint row, col;
  gint x1, y1, x2, y2;
  gint cnt;
  gint has_alpha, ind;

  /*
   *  Get the size of the input image. (This will/must be the same
   *  as the size of the output image.  Also get alpha info.
   */
  width = drawable->width;
  height = drawable->height;
  bytes = drawable->bpp;
  has_alpha = gimp_drawable_has_alpha(drawable->drawable_id);
  /*
   *  allocate row buffers
   */
  prev_row = g_new (guchar, (x2 - x1 + 2) * bytes);
  cur_row  = g_new (guchar, (x2 - x1 + 2) * bytes);
  next_row = g_new (guchar, (x2 - x1 + 2) * bytes);
  dest     = g_new (guchar, (x2 - x1) * bytes);

/*
  pr = prev_row + bytes;
  cr = cur_row + bytes;
  nr = next_row + bytes;
*/

  pr = prev_row + bytes;
  cr = cur_row + bytes;
  nr = next_row + bytes;


  for (cnt = 1; cnt <= pivals.blur_rcount; cnt++)
    {
      /*
       *  prepare the first row and previous row
       */


      /*
       *  loop through the rows, applying the selected convolution
       */
      for (row = 0; row < ysize; row++)
	{
	  /*  prepare the next row the first row and previous row */
		if (row == 0) pr = oscan[row] else pr = oscan[row-1] ; 
		cr = oscan[row];
		if (row == ysize-1)   nr = oscan[row];

	  d = nscan[row];
	  ind = 0;
	  for (col = 0; col < (xsize) * bytes; col++)
	    {
	      if (1)  //everytime.... eventually get rid of this if
		{
		  ind++;
		  if (ind==bytes || !(has_alpha))
		    {
		      /*
		       *  If no alpha channel,
		       *   or if there is one and this is it...
		       */
		      *d++ = ((int) pr[col - bytes] + (int) pr[col] +
			      (int) pr[col + bytes] +
			      (int) cr[col - bytes] + (int) cr[col] +
			      (int) cr[col + bytes] +
			      (int) nr[col - bytes] + (int) nr[col] +
			      (int) nr[col + bytes]) / 9;
		      ind = 0;
                    }
		  else
		    {
		      /*
		       *  otherwise we have an alpha channel,
		       *   but this is a color channel
		       */
		      *d++ = ((int)
			      (((double) (pr[col - bytes] * pr[col - ind]) 
				+ (double) (pr[col] * pr[col + bytes - ind])
				+ (double) (pr[col + bytes] * pr[col + 2*bytes - ind])
				+ (double) (cr[col - bytes] * cr[col - ind])
				+ (double) (cr[col] * cr[col + bytes - ind])
				+ (double) (cr[col + bytes] * cr[col + 2*bytes - ind])
				+ (double) (nr[col - bytes] * nr[col - ind])
				+ (double) (nr[col] * nr[col + bytes - ind]) 
				+ (double) (nr[col + bytes] * nr[col + 2*bytes - ind]))
                             / ((double) pr[col - ind]
				+ (double) pr[col + bytes - ind]
				+ (double) pr[col + 2*bytes - ind]
				+ (double) cr[col - ind]
				+ (double) cr[col + bytes - ind]
				+ (double) cr[col + 2*bytes - ind]
				+ (double) nr[col - ind]
				+ (double) nr[col + bytes - ind]
				+ (double) nr[col + 2*bytes - ind])));
                    }
		  /*
		   *  Otherwise, this pixel was not selected for randomization,
		   *  so use the current value.
		   */
                }
	      else
		{
		  *d++ = (int) cr[col];
                }
            }

        }
    }
}


/////////////////////


static void
equ_histogram(long hist[3][256], unsigned char lut[3][256], long channels, long width, long height)
{
	int    i, k, j;
	int    part[3][257]; /* Partition */
	double pixels_per_value;
	double desired;

	/* Calculate partial sums */

	for (k = 0; k < channels; k++)
		for (i = 1; i < 256; i++)
			hist[k][i] += hist[k][i - 1];

	/* Find partition points */

	pixels_per_value = (double) (width * height) / 256;

	for (k = 0; k < channels; k++) {
		/* First and last points in partition */

		part[k][0]   = 0;
		part[k][256] = 256;

		/* Find intermediate points */

		j = 0;
		
		for (i = 1; i < 256; i++) {
			desired = i * pixels_per_value;
			
			while (hist[k][j + 1] <= desired)
				j++;

			/* Nearest sum */

			if ((desired - hist[k][j]) < (hist[k][j + 1] - desired))
				part[k][i] = j;
			else
				part[k][i] = j + 1;
		} /* for */
	} /* for */

	/* Create equalization LUT */

	for (k = 0; k < channels; k++)
		for (j = 0; j < 256; j++) {
			i = 0;

			while (part[k][i + 1] <= j)
				i++;

			lut[k][j] = i;
		} /* for */
} /* equ_histogram */


static void
do_equalize(uchar **oscan, uchar **nscan, int xsize, int ysize, int channels)
{
	uchar *src, *dest, old, new;
	long           hist[3][256]; /* Histogram */
	uchar  lut[3][256];
	int            x1, y1, x2, y2;
	long          i, x, y, off;
	int            k;


	switch(channels){
	case 3:
	case 4: 
		break;
	default:
		sysfatal("can't handle  %d-channel images, need 24(32)-bpp depth", channels);	
	}

	/* Initialize histogram */

	for (i = 0; i < 256; i++)
		for (k = 0; k < channels; k++)
			hist[k][i] = 0;

	/* Build histogram */


	for (x = 0; x < xsize; x++) {
		for (k = 0; k < channels; k++)
		
		for(y=0; y<ysize; y++){  // for every point in scanline
			off=channels*x+k;
			old =  oscan[y][off];


				hist[k][old]++;
		}

	} 
	/* Build equalization LUT */
	
	equ_histogram(hist, lut, channels, xsize, ysize);


	/* Substitute */


	for (x = 0; x < xsize; x++) {
		for (k = 0; k < channels; k++) {
			for (y = 0; y < ysize; y++) {
				off=channels*x+k;
				old =  oscan[y][off];
				if (k != ALPHA)
					new = lut[k][old];
				else
					new = old;
				nscan[y][off] =  new;
			}
		}
	} 
} /* do_equalize */




int
max(int a, int b)
{
	if(a > b)
		return a;
	return b;
}


Memimage*
mknewimg(int xsize, int ysize, Memimage *m)
{
	int i, j, bpl, nchan;
	Memimage *new;
	uchar **oscan, **nscan;

	new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
	if(new == nil)
		sysfatal("can't allocate new image: %r");

	oscan = malloc(Dy(m->r)*sizeof(uchar*));
	nscan = malloc(max(ysize, Dy(m->r))*sizeof(uchar*));
	if(oscan == nil || nscan == nil)
		sysfatal("can't allocate: %r");

	/* unload original image into scan lines */
	bpl = bytesperline(m->r, m->depth);
	for(i=0; i<Dy(m->r); i++){
		oscan[i] = malloc(bpl);
		if(oscan[i] == nil)
			sysfatal("can't allocate: %r");
		j = unloadmemimage(m, Rect(m->r.min.x, m->r.min.y+i, m->r.max.x, m->r.min.y+i+1), oscan[i], bpl);
		if(j != bpl)
			sysfatal("unloadmemimage");
	}

	/* allocate scan lines for destination. we do y first, so need at least Dy(m->r) lines */
	bpl = bytesperline(Rect(0, 0, xsize, Dy(m->r)), m->depth);
	for(i=0; i<max(ysize, Dy(m->r)); i++){
		nscan[i] = malloc(bpl);
		if(nscan[i] == nil)
			sysfatal("can't allocate: %r");
	}

	nchan = m->depth/8;

	do_equalize(oscan, nscan, xsize, ysize, nchan);



	/* pack data into destination */
	bpl = bytesperline(new->r, m->depth);
	for(i=0; i<ysize; i++){
		j = loadmemimage(new, Rect(0, i, xsize, i+1), nscan[i], bpl);
		if(j != bpl)
			sysfatal("loadmemimage: %r");
	}
	return new;
}


void
main(int argc, char *argv[])
{
	int i, fd, xsize, ysize, dummy;
	Rectangle rparam;
	Memimage *m, *new, *t1, *t2;
	char *file;
	ulong tchan;
	char tmp[100];
	double v;

	memimageinit();
	memset(&rparam, 0, sizeof rparam);
	xsize = ysize = 0;


	file = "<stdin>";
	fd = 0;

	switch (argc) {
	case 2:
		file = argv[1];
		fd = open(file, OREAD);
		if(fd < 0)
			sysfatal("can't open %s: %r", file);
	case 1:
		break;
	default:
		usage();
	}

	m = readmemimage(fd);
	if(m == nil)
		sysfatal("can't read %s: %r", file);

	xsize = Dx(m->r);
	ysize = Dy(m->r);

	new = nil;
	switch(m->chan){

	case RGB24:
	case RGBA32:
	case ARGB32:
	case XRGB32:
		new = mknewimg(xsize, ysize, m);
		break;

	case GREY1:
	case GREY2:
	case GREY4:
	case GREY8:
	case CMAP8:
	case RGB15:
	case RGB16:
		tchan = RGB24;
		goto Convert;

	Convert:
		/* use library to convert to byte-per-chan form, then convert back */
		t1 = allocmemimage(m->r, tchan);
		if(t1 == nil)
			sysfatal("can't allocate temporary image: %r");
		memimagedraw(t1, t1->r, m, m->r.min, nil, ZP);
		t2 = mknewimg(xsize, ysize, t1);
		freememimage(t1);
		new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
		if(new == nil)
			sysfatal("can't allocate new image: %r");
		/* should do error diffusion here */
		memimagedraw(new, new->r, t2, t2->r.min, nil, ZP);
		freememimage(t2);
		break;

	default:
		sysfatal("can't handle channel type %s", chantostr(tmp, m->chan));
	}

	assert(new);
	if(writememimage(1, new) < 0)
		sysfatal("write error on output: %r");

	exits(nil);
}
